// For general information on the Pynini grammar compilation library, see
// pynini.opengrm.org.

#ifndef REPEAT_H_
#define REPEAT_H_

#include <memory>

#include <fst/fstlib.h>

namespace fst {

// This function is a generalization of FST closure and Perl's curly brace
// quantifiers. It destructively computes the concatenative closure of an input
// FST as follows. If A transduces strings x to y with weight w, then
// Repeat(A, 0, 0) is equivalent to Closure(A, CLOSURE_STAR) which mutates A so
// that it transduces between empty strings with weight One, transduces strings
// x to y with weight w, transduces xx to yy with weight w \otimes w, strings
// xxx to yyy with weight w \otimes w \otimes w (and so on).
//
// When called with two non-zero positive integers as the trailing arguments,
// these act as upper and lower bounds, respectively, for the number of cycles
// through the original FST one is permitted to take in the modified FST. So,
// Repeat(A, 0, 1) is mutates A so it transduces between empty strings with
// weight One and // transduces strings x to y with weight w, similar to the ?
// quantifier in Perl. And, Repeat(A, 2, 5) mutates A so that it behaves like
// the concatenation of between 2 and 5 A's.
//
// When the third argument is zero, it is interpreted to indicate an infinite
// upper bound. Thus, Repeat(A, 1, 0) is equivalent to Closure(A, CLOSURE_PLUS).
//
// The following provide equvialents to the Perl operators:
//
//     /x*/        Repeat(x, 0, 0)
//     /x+/        Repeat(x, 1, 0)
//     /x?/        Repeat(x, 0, 1)
//     /x{N}/      Repeat(x, N, N)
//     /x{M,N}/    Repeat(x, M, N)
//     /x{N,}/     Repeat(x, N, 0)
//     /x{,N}/     Repeat(x, 0, N)
//
template <class Arc>
void Repeat(MutableFst<Arc> *fst, int lower, int upper) {
  if (upper <= 0) {
    // Special cases.
    if (lower <= 0) {
      Closure(fst, CLOSURE_STAR);
      return;
    } else if (lower == 1) {
      Closure(fst, CLOSURE_PLUS);
      return;
    }
  }
  // Generic cases.
  typename Arc::Weight One = Arc::Weight::One();
  if (lower <= 0)   // Lower bound includes 0 repetitions.
    fst->SetFinal(fst->Start(), One);
  std::unique_ptr<MutableFst<Arc>> tfst(fst->Copy());
  for (int i = 0; i < lower - 1; ++i)
    Concat(fst, *tfst);
  if (upper <= 0) {  // Upper bound is infinite.
    // Concatenates a *-ed copy.
    Closure(tfst.get(), CLOSURE_STAR);
    Concat(fst, *tfst);
  } else {           // Upper bound is finite.
    // Concatenates ?-ed copies.
    tfst->SetFinal(tfst->Start(), One);
    for (int i = lower; i < upper; ++i)
      Concat(fst, *tfst);
  }
}

}  // namespace fst

#endif  // REPEAT_H_

